# Spec version 0.3.1
# This file is auto-generated by setup.py

from .thing_types import ThingTypesBase


class AnnotationTypesBase(ThingTypesBase):
    """
    Inherit from this class to build your own custom annotation
    vocabularies. Each member's name should be the name of the
    type, and each member's value should be the URI of that type.
    """
    ...


class AnnotationTypes(AnnotationTypesBase):
    """
    This enum contains the URIs for the MMIF annotation types defined in the spec version 0.3.1.
    Use this to quickly get the correct URIs for those types, or use the objects themselves in your
    code and they will serialize to the strings.
    """
    Annotation = 'http://mmif.clams.ai/0.3.1/vocabulary/Annotation'
    Region = 'http://mmif.clams.ai/0.3.1/vocabulary/Region'
    TimePoint = 'http://mmif.clams.ai/0.3.1/vocabulary/TimePoint'
    Interval = 'http://mmif.clams.ai/0.3.1/vocabulary/Interval'
    Span = 'http://mmif.clams.ai/0.3.1/vocabulary/Span'
    TimeFrame = 'http://mmif.clams.ai/0.3.1/vocabulary/TimeFrame'
    Chapter = 'http://mmif.clams.ai/0.3.1/vocabulary/Chapter'
    Polygon = 'http://mmif.clams.ai/0.3.1/vocabulary/Polygon'
    BoundingBox = 'http://mmif.clams.ai/0.3.1/vocabulary/BoundingBox'
    VideoObject = 'http://mmif.clams.ai/0.3.1/vocabulary/VideoObject'
    Relation = 'http://mmif.clams.ai/0.3.1/vocabulary/Relation'
    Alignment = 'http://mmif.clams.ai/0.3.1/vocabulary/Alignment'
