# This file is auto-generated by setup.py

from enum import Enum


class ThingTypesBase(Enum):
    def _serialize(self):
        return str(self)

    def __str__(self):
        return self.value


class ThingType(ThingTypesBase):
    """
    This enum contains the URI for the MMIF Thing types defined in the spec version 0.3.1.
    Use this to quickly get the correct URI for this type, or use the object itself in your
    code and it will serialize to the string.
    """
    Thing = 'http://mmif.clams.ai/0.3.1/vocabulary/Thing'
